/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/* $Source: src/usr/diag/prdf/common/framework/service/prdfRasServices_common.C $ */
/*                                                                        */
/* OpenPOWER HostBoot Project                                             */
/*                                                                        */
/* Contributors Listed Below - COPYRIGHT 2013,2015                        */
/* [+] International Business Machines Corp.                              */
/*                                                                        */
/*                                                                        */
/* Licensed under the Apache License, Version 2.0 (the "License");        */
/* you may not use this file except in compliance with the License.       */
/* You may obtain a copy of the License at                                */
/*                                                                        */
/*     http://www.apache.org/licenses/LICENSE-2.0                         */
/*                                                                        */
/* Unless required by applicable law or agreed to in writing, software    */
/* distributed under the License is distributed on an "AS IS" BASIS,      */
/* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or        */
/* implied. See the License for the specific language governing           */
/* permissions and limitations under the License.                         */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */

/** @file  prdfRasServices_common.C
 *  @brief Utility code to parse an SDC and produce the appropriate error log.
 */

#include <prdfRasServices.H>
#include <prdfPfa5Data.h>
#include <time.h>
#include <iipServiceDataCollector.h>
#include <prdf_service_codes.H>
#include <prdfGlobal.H>
#include <prdfErrlUtil.H>
#include <prdfCallouts.H>
#include <prdfMemoryMru.H>
#include <prdfPlatServices.H>

// For compression routines
#define PRDF_COMPRESSBUFFER_COMPRESS_FUNCTIONS
#include <prdfCompressBuffer.H>

#include <utilmem.H> //For UtilMem stream class (outputting PfaData).
#include <utilfile.H>
#include <vector>
#include <algorithm>
#include <iipSystem.h>         //For RemoveStoppedChips

#ifdef __HOSTBOOT_MODULE

  #include <stdio.h>

  #ifdef __HOSTBOOT_RUNTIME
    #include <prdfCenMbaDynMemDealloc_rt.H>
  #endif

#else
  #include <srcisrc.H>
  #include <utilreg.H> //For registry functions
  #include <evenmgt.H>
  #include <rmgrBaseClientLib.H>  //for rmgrSyncFile
  #include <prdfSdcFileControl.H>
  #include <prdfCenMbaDynMemDealloc_rt.H>
#endif

using namespace TARGETING;

namespace PRDF
{

using namespace PlatServices;

//------------------------------------------------------------------------------
// Local Globals
//------------------------------------------------------------------------------

RasServices thisServiceGenerator;

//------------------------------------------------------------------------------

ServiceGeneratorClass & ServiceGeneratorClass::ThisServiceGenerator(void)
{
  return thisServiceGenerator;
}

//------------------------------------------------------------------------------

void ErrDataService::Initialize()
{
    iv_serviceActionCounter = 0;
}

//------------------------------------------------------------------------------

errlHndl_t ErrDataService::GenerateSrcPfa( ATTENTION_TYPE i_attnType,
                                           ServiceDataCollector & io_sdc,
                                           bool & o_initiateHwudump,
                                           TargetHandle_t & o_dumpTrgt,
                                           errlHndl_t & o_dumpErrl,
                                           uint32_t & o_dumpErrlActions)
{
    #define PRDF_FUNC "[ErrDataService::GenerateSrcPfa] "

    o_initiateHwudump = false;
    o_dumpTrgt        = NULL;
    o_dumpErrl        = NULL;
    o_dumpErrlActions = 0;

#ifdef __HOSTBOOT_MODULE
    using namespace ERRORLOG;
    using namespace HWAS;
    errlSeverity_t severityParm = ERRL_SEV_RECOVERED;
#else
    uint8_t sdcSaveFlags = SDC_NO_SAVE_FLAGS;
    size_t  sz_uint8    = sizeof(uint8_t);
    uint8_t sdcBuffer[sdcBufferSize];  //buffer to use for SDC flatten
    errlSeverity severityParm = ERRL_SEV_RECOVERED;
#endif

    PRDcallout thiscallout;
    PRDpriority thispriority;
    epubProcedureID thisProcedureID;

    // Init Action Parm to most common usage, Service Action Required and
    // Report Externally. Note this is like the old Signal
    // Event: OS Viewable (Parable) or OS Hidden.
    // Also set the Call Home Flag. This should be set when IBM Service is required.
    // For PRD this is for UnRecoverable and Predictive errors. Setting it here will
    // take care of this. The Hidden and Informational cases will reassign the actionFlag.
    uint32_t actionFlag = (ERRL_ACTION_SA | ERRL_ACTION_REPORT | ERRL_ACTION_CALL_HOME);

    HWSV::hwsvDiagUpdate l_diagUpdate = HWSV::HWSV_DIAG_NEEDED;

    GardAction::ErrorType prdGardErrType;
    HWAS::GARD_ErrorType gardErrType = HWAS::GARD_NULL;
    HWAS::DeconfigEnum deconfigState = HWAS::NO_DECONFIG;

    bool ForceTerminate = false;
    bool iplDiagMode = false;
    bool deferDeconfig = false;

    ++iv_serviceActionCounter;

    uint16_t PRD_Reason_Code = 0;

    //**************************************************************
    // Initial set up by Attention Type
    //**************************************************************

    ////////////////////////////////////////////////////////////////
    // Machine Check ATTN (CHECKSTOP)
    ////////////////////////////////////////////////////////////////
    if (i_attnType == MACHINE_CHECK)
    {
        #ifdef __HOSTBOOT_MODULE

        // Do nothing. Saved SDCs only supported on FSP.

        #ifdef __HOSTBOOT_RUNTIME
        PRDF_ERR( PRDF_FUNC "HBRT should NOT have any CS attns!" );
        #endif

        #else

        severityParm = ERRL_SEV_UNRECOVERABLE;

        // Check for SUE-CS condition flags.
        if ((!io_sdc.IsUERE() ) &&
            ( io_sdc.IsSUE()  )   )
        {
            //Read current SDC state flags from registry
            errlHndl_t errorLog = UtilReg::read ( "prdf/RasServices",
                                                  &sdcSaveFlags, sz_uint8 );
            if (errorLog)
            {
                PRDF_ERR( PRDF_FUNC "Failure in SDC flag Registry read" );
                PRDF_COMMIT_ERRL(errorLog, ERRL_ACTION_REPORT);
            }
            else if (sdcSaveFlags & SDC_SAVE_UE_FLAG) //check if UE log stored
            {                                         // then use it.
                bool l_rc = SdcRetrieve(SDC_SAVE_UE_FLAG, sdcBuffer);
                if (l_rc)
                {
                    PRDF_ERR( PRDF_FUNC "Failure in UE SDC Retrieve Function" );
                }
                else
                {
                    ServiceDataCollector origSdc = io_sdc; // Save original SDC
                    io_sdc = sdcBuffer;                    // Set the new SDC

                    // Add the original signature to the new SDC's
                    // multi-signature list.
                    io_sdc.AddSignatureList( *(origSdc.GetErrorSignature()) );

                    // This is a checkstop attention so some values will need
                    // to be overridden.

                    io_sdc.setPrimaryAttnType( origSdc.getPrimaryAttnType() );

                    if ( origSdc.Terminate() )
                        io_sdc.SetFlag(ServiceDataCollector::TERMINATE);
                }
            }
            else if (sdcSaveFlags & SDC_SAVE_SUE_FLAG ) //else check if SUE log
            {                                           // stored then use it.
                bool l_rc = SdcRetrieve(SDC_SAVE_SUE_FLAG, sdcBuffer);
                if (l_rc)
                {
                    PRDF_ERR( PRDF_FUNC "Failure in SUE SDC Retrieve Function" );
                }
                else
                {
                    ServiceDataCollector origSdc = io_sdc; // Save original SDC
                    io_sdc = sdcBuffer;                    // Set the new SDC

                    // Add the original signature to the new SDC's
                    // multi-signature list.
                    io_sdc.AddSignatureList( *(origSdc.GetErrorSignature()) );

                    // This is a checkstop attention so some values will need
                    // to be overridden.

                    io_sdc.setPrimaryAttnType( origSdc.getPrimaryAttnType() );

                    if ( origSdc.Terminate() )
                        io_sdc.SetFlag(ServiceDataCollector::TERMINATE);
                }
            }
        }

        #endif  // if not __HOSTBOOT_HOSTBOOT
    }
    ////////////////////////////////////////////////////////////////
    // Recoverable ATTN or Unit CheckStop
    ////////////////////////////////////////////////////////////////
    else if (i_attnType == RECOVERABLE  || i_attnType == UNIT_CS )
    {
#ifndef  __HOSTBOOT_MODULE
        if //Ue-Re RECOVERABLE condition.
            ((io_sdc.IsUERE()   ) &&
             (!io_sdc.IsSUE()   ) &&
             (!io_sdc.IsUsingSavedSdc() ) )  // Don't save File if we are Re-Syncing an SDC
        {
            bool l_rc = SdcSave(SDC_SAVE_UE_FLAG, io_sdc);
            if (l_rc)
            {
                PRDF_ERR( PRDF_FUNC "Failure in UE SDC Save Function" );
            }
        }
        else if //Sue-Re RECOVERABLE condition.
            ((!io_sdc.IsUERE()   ) &&
             (io_sdc.IsSUE()     ) &&
             (!io_sdc.IsUsingSavedSdc() ) )  // Don't save File if we are Re-Syncing an SDC

        {
            bool l_rc = SdcSave(SDC_SAVE_SUE_FLAG, io_sdc);
            if (l_rc)
            {
                PRDF_ERR( PRDF_FUNC "Failure in SUE SDC Save Function" );
            }
        }
#endif  // if not __HOSTBOOT_MODULE

        if (!io_sdc.IsLogging() )
        {
            // This is a Hidden Log
            severityParm = ERRL_SEV_INFORMATIONAL;
            actionFlag = (actionFlag | ERRL_ACTION_HIDDEN);
        }
        else if ( io_sdc.IsServiceCall() ) //At Thresold
        {
            severityParm = ERRL_SEV_PREDICTIVE;
        }
        else  //Recovered
        {
            severityParm = ERRL_SEV_RECOVERED;
            //  Recovered error should be Hidden, and No Service Action
            actionFlag = ERRL_ACTION_HIDDEN;
        }
    }
    ////////////////////////////////////////////////////////////////
    // Special ATTN
    ////////////////////////////////////////////////////////////////
    else if (i_attnType == SPECIAL)
    {
        if (io_sdc.IsServiceCall())
        //Bit Steered already, or Bit Steer Not supported
        {
            severityParm = ERRL_SEV_PREDICTIVE;
        }
        else if ( (!io_sdc.IsServiceCall()) && (!io_sdc.IsLogging()) ) // Special Attn Clean Up
        {
            severityParm = ERRL_SEV_INFORMATIONAL;
            //Hidden, No Service Action for Infomational
            actionFlag = ERRL_ACTION_HIDDEN;
        }
        else if ( (!io_sdc.IsServiceCall()) && (io_sdc.IsLogging()) ) // Special Attn Bit Steer Normal Condition
        {
            severityParm = ERRL_SEV_RECOVERED;
            //Hidden, No Service Action for Recovered
            actionFlag = ERRL_ACTION_HIDDEN;
        }
    }

    //**************************************************************
    // Set Gard Error Type and state
    //**************************************************************

    prdGardErrType = io_sdc.QueryGard();
    switch ( prdGardErrType )
    {
        case GardAction::NoGard:     gardErrType = HWAS::GARD_NULL;       break;
        case GardAction::Predictive: gardErrType = HWAS::GARD_Predictive; break;
        case GardAction::Fatal:      gardErrType = HWAS::GARD_Fatal;      break;
        default:
            gardErrType = HWAS::GARD_NULL;
    }

    //**************************************************************
    // Callout loop to set up Reason code and SRC word 9
    //**************************************************************

    // Must go thru callout list to look for RIOPORT procedure callouts,
    // since they require the port info to be in SRC Word 9
    bool HW = false;
    bool SW = false;
    bool SW_High = false;
    bool SecondLevel = false;
    uint32_t SrcWord7 = 0;
    uint32_t SrcWord9 = 0;

    // Should not gard hardware if there is a hardware callout at LOW priority
    // and a symbolic FRU indicating a possibility of a software error at MED or
    // HIGH priority.
    bool sappSwNoGardReq = false, sappHwNoGardReq = false;

    const SDC_MRU_LIST & mruList = io_sdc.getMruList();
    int32_t calloutsPlusDimms = mruList.size();

    for ( SDC_MRU_LIST::const_iterator it = mruList.begin();
          it < mruList.end(); ++it )
    {
        thiscallout = it->callout;
        if ( PRDcalloutData::TYPE_SYMFRU == thiscallout.getType() )
        {
            if ( (EPUB_PRC_SP_CODE   == thiscallout.flatten()) ||
                 (EPUB_PRC_PHYP_CODE == thiscallout.flatten()) )
            {
                SW = true;

                if ( MRU_LOW != it->priority )
                {
                    sappSwNoGardReq = true;
                }

                if ( MRU_MED == it->priority )
                {
                    SW_High = true;
                }
            }
            else if ( EPUB_PRC_LVL_SUPP == thiscallout.flatten())
            {
                SecondLevel = true;

                if ( MRU_LOW != it->priority )
                {
                    sappSwNoGardReq = true;
                }
            }
        }
        else if ( PRDcalloutData::TYPE_MEMMRU == thiscallout.getType() )
        {
            MemoryMru memMru (thiscallout.flatten());
            SrcWord9 = memMru.toUint32(); // Get MemMru value

            TargetHandleList partList = memMru.getCalloutList();
            uint32_t partCount = partList.size();

            calloutsPlusDimms = calloutsPlusDimms + partCount -1;
            HW = true; //hardware callout

            if ( MRU_LOW == it->priority )
            {
                sappHwNoGardReq = true;
            }
        }
        else // PRDcalloutData::TYPE_TARGET
        {
            HW = true; // Hardware callout

            // Determines if all the hardware callouts have low priority.

            if ( MRU_LOW == it->priority )
            {
                sappHwNoGardReq = true;
            }
        }
    }

    ////////////////////////////////////////////////////////////////
    //Set the PRD Reason Code based on the flags set in the above callout loop.
    ////////////////////////////////////////////////////////////////

    if (HW == true && SW == true)
    {
        if (SW_High == true)
            PRD_Reason_Code = PRDF_DETECTED_FAIL_SOFTWARE_PROBABLE;
        else
            PRD_Reason_Code = PRDF_DETECTED_FAIL_HARDWARE_PROBABLE;
    }
    else if (HW == true && SW == false && SecondLevel == true)
        PRD_Reason_Code = PRDF_DETECTED_FAIL_HARDWARE_PROBABLE;
    else if (HW == true && SW == false && SecondLevel == false)
        PRD_Reason_Code = PRDF_DETECTED_FAIL_HARDWARE;
    else if (HW == false && SW == true)
        PRD_Reason_Code = PRDF_DETECTED_FAIL_SOFTWARE;
    else
    {
        // If we get here both HW and SW flags were false. Callout may be
        // Second Level Support only, or a procedure not checked in the SW
        // flag code.
        PRD_Reason_Code = PRDF_DETECTED_FAIL_HARDWARE_PROBABLE;
    }

    SrcWord7  = io_sdc.getPrimaryAttnType() << 8;
    SrcWord7 |= io_sdc.getSecondaryAttnType();

    //--------------------------------------------------------------------------
    // Check for IPL Diag Mode and set up for Deferred Deconfig
    //--------------------------------------------------------------------------

    #if defined(__HOSTBOOT_MODULE) && !defined(__HOSTBOOT_RUNTIME)

    iplDiagMode = PlatServices::isInMdiaMode();

    // Deferred Deconfig should be used throughout all of Hostboot (both
    // checkForIplAttns() and MDIA).
    if ( HWAS::GARD_NULL != gardErrType )
    {
        deferDeconfig = true;
        deconfigState = HWAS::DECONFIG;
    }

    #endif

    //**************************************************************
    // Update Error Log with SRC
    //**************************************************************
    ErrorSignature * esig = io_sdc.GetErrorSignature();

    updateSrc( esig->getChipId(), SrcWord7, esig->getSigId(),
               SrcWord9, PRD_Reason_Code);

    //**************************************************************
    //  Add SDC Capture data to Error Log User Data here only if
    //    there are 4 or more callouts,
    //    (including Dimm callouts in the MemoryMru).
    //**************************************************************
    bool capDataAdded = false;
    if (calloutsPlusDimms > 3)
    {
        AddCapData( io_sdc.GetCaptureData(),    iv_errl );
        AddCapData( io_sdc.getTraceArrayData(), iv_errl );
        capDataAdded = true;
    }

    // make sure serviceAction doesn't override errl severity
    iv_errl->setSev(severityParm);

    if (ERRL_ACTION_HIDDEN == actionFlag)
    {  // Diagnostics is not needed in the next IPL cycle for non-visible logs
        l_diagUpdate = HWSV::HWSV_DIAG_NOT_NEEDED;
    }

    //**************************************************************
    // Add each mru/callout to the error log.
    //**************************************************************

    // Change deconfigState only based on Gard Types.
    // This only affects FSP code since Hostboot macro is no-op.
    // This is needed for FSP Reconfig Loop to work.
    if ( HWAS::GARD_NULL != gardErrType )
    {
        PRDF_RECONFIG_LOOP(deconfigState);
    }

    // OPAL has requested that we not gard any hardware if it is possible that
    // software generated the attention. This will be determined if there is a
    // software callout with higher priority than a hardware callout.
    if ( sappSwNoGardReq && sappHwNoGardReq &&          // Gard requirements met
         isHyprConfigOpal() && isHyprRunning() &&       // OPAL is running
         !isMfgAvpEnabled() && !isMfgHdatAvpEnabled() ) // No AVPs running
    {
        gardErrType = HWAS::GARD_NULL;
        prdGardErrType = GardAction::NoGard;
    }

    for ( SDC_MRU_LIST::const_iterator it = mruList.begin();
          it < mruList.end(); ++it )
    {
        thispriority = it->priority;
        thiscallout = it->callout;
        HWAS::GARD_ErrorType tmpGard = gardErrType;
        HWAS::DeconfigEnum tmpDeconfig = deconfigState;

        if( HWAS::GARD_NULL != gardErrType )
        {
            if( NO_GARD == it->gardState )
            {
                tmpGard = HWAS::GARD_NULL;
                tmpDeconfig = HWAS::NO_DECONFIG;
            }
        }

        if( PRDcalloutData::TYPE_TARGET == thiscallout.getType() )
        {
            PRDF_HW_ADD_CALLOUT(thiscallout.getTarget(),
                                thispriority,
                                tmpDeconfig,
                                iv_errl,
                                tmpGard,
                                severityParm,
                                l_diagUpdate);
        }
        else if(PRDcalloutData::TYPE_PROCCLK == thiscallout.getType() ||
                PRDcalloutData::TYPE_PCICLK  == thiscallout.getType())
        {
            PRDF_ADD_CLOCK_CALLOUT(iv_errl,
                                   thiscallout.getTarget(),
                                   thiscallout.getType(),
                                   thispriority,
                                   tmpDeconfig,
                                   tmpGard);
        }
        else if ( PRDcalloutData::TYPE_MEMMRU == thiscallout.getType() )
        {
            MemoryMru memMru (thiscallout.flatten());

            TargetHandleList partList = memMru.getCalloutList();
            for ( TargetHandleList::iterator it = partList.begin();
                  it != partList.end(); it++ )
            {
                PRDF_HW_ADD_CALLOUT( *it,
                                     thispriority,
                                     tmpDeconfig,
                                     iv_errl,
                                     tmpGard,
                                     severityParm,
                                     l_diagUpdate );
            }
        }
        else if ( PRDcalloutData::TYPE_SYMFRU == thiscallout.getType() )
        {
            thisProcedureID = epubProcedureID(thiscallout.flatten());

            PRDF_DTRAC( PRDF_FUNC "thisProcedureID: %x, thispriority: %x, severityParm: %x",
                   thisProcedureID, thispriority,severityParm );

            PRDF_HW_ADD_PROC_CALLOUT(thisProcedureID,
                                     thispriority,
                                     iv_errl,
                                     severityParm);

            // Use the flags set earlier to determine if the callout is just Software (SP code or Phyp Code).
            // Add a Second Level Support procedure callout Low, for this case.
            if (HW == false && SW == true && SecondLevel == false)
            {
                PRDF_DTRAC( PRDF_FUNC "thisProcedureID= %x, thispriority=%x, severityParm=%x",
                   EPUB_PRC_LVL_SUPP, MRU_LOW, severityParm );

                PRDF_HW_ADD_PROC_CALLOUT(EPUB_PRC_LVL_SUPP,
                                         MRU_LOW,
                                         iv_errl,
                                         severityParm);

                SecondLevel = true;
            }

        }
    }

    // Send the dynamic memory Dealloc message for DIMMS for Predictive
    // callouts.
    // We can not check for ERRL severity here as there are some cases
    // e.g. DD02 where we create a Predictive error log but callouts
    // are not predictive.
    if ( HWAS::GARD_Predictive == gardErrType )
    {
        deallocateDimms( mruList );
    }

    //**************************************************************
    // Build Dump Flags and PFA5 data
    //**************************************************************

    PfaData pfaData;

    initPfaData( io_sdc, i_attnType, deferDeconfig, actionFlag, severityParm,
                 prdGardErrType, pfaData, o_dumpTrgt );

    //**************************************************************
    // Check for Terminating the system for non mnfg conditions.
    //**************************************************************

    ForceTerminate = checkForceTerm( io_sdc, o_dumpTrgt, pfaData );

    //*************************************************************
    // Check for Manufacturing Mode terminate here and then do
    // the needed overrides on ForceTerminate flag.
    //*************************************************************
    if ( PlatServices::mnfgTerminate() && !ForceTerminate )
    {
        ForceTerminate = true;
        if ( !((severityParm == ERRL_SEV_RECOVERED) ||
               (severityParm == ERRL_SEV_INFORMATIONAL)) &&
             iplDiagMode  &&
             !HW )
        {
            //Terminate in Manufacturing Mode, in IPL mode, for visible log, with no HW callouts.
            PRDF_SRC_WRITE_TERM_STATE_ON(iv_errl, SRCI_TERM_STATE_MNFG);
        }
        // Do not terminate if recoverable or informational.
        // Do not terminate if deferred deconfig.
        else if ( deferDeconfig                            ||
                  (severityParm == ERRL_SEV_RECOVERED    ) ||
                  (severityParm == ERRL_SEV_INFORMATIONAL)  )
        {
            ForceTerminate = false;
            actionFlag = (actionFlag | ERRL_ACTION_DONT_TERMINATE);
        }
        else
        {
            PRDF_SRC_WRITE_TERM_STATE_ON(iv_errl, SRCI_TERM_STATE_MNFG);
        }

        pfaData.errlActions = actionFlag;
    }

    // Needed to move the errl add user data sections here because of some updates
    // of the data required in the Aysnc section for the SMA dual reporting fix.

    //**************************************************************
    // Add the PFA data to Error Log User Data
    //**************************************************************
    UtilMem l_membuf;
    l_membuf << pfaData;
    PRDF_ADD_FFDC( iv_errl, (const char*)l_membuf.base(), l_membuf.size(),
                   ErrlVer1, ErrlSectPFA5_1 );

    //**************************************************************
    // Add SDC Capture data to Error Log User Data
    //**************************************************************
    // Pulled some code out to incorporate into AddCapData
    // Check to make sure Capture Data wasn't added earlier.
    if (!capDataAdded)
    {
        AddCapData( io_sdc.GetCaptureData(),    iv_errl );
        AddCapData( io_sdc.getTraceArrayData(), iv_errl );
    }

    // Note moved the code from here, that was associated with checking for the last
    // functional core to be before the PFA data is placed in error log.

    // Collect PRD trace
    // NOTE: Each line of trace is on average 36 bytes so 768 bytes should get
    //       us around 21 lines of trace output.
    PRDF_COLLECT_TRACE(iv_errl, 768);

    //**************************************************************
    // Commit the error log.
    // This will also perform Gard and Deconfig actions.
    // Do the Unit Dumps if needed.
    //**************************************************************

    if ( io_sdc.IsDontCommitErrl() && !io_sdc.IsUnitCS() &&
         (MACHINE_CHECK != pfaData.priAttnType) )
    {
        // User did not want this error log committed, so delete it.
        delete iv_errl; iv_errl = NULL;
    }
    else
    {
        // Add the MNFG trace information.
        MnfgTrace( io_sdc.GetErrorSignature(), pfaData );

        // If this is not a terminating condition, commit the error log. If the
        // error log is not committed, the error log will be passed back to
        // PRDF::main() and eventually ATTN.
        if ( MACHINE_CHECK != pfaData.priAttnType && !ForceTerminate &&
             !pfaData.TERMINATE )
        {
            // Handle any unit checkstop conditions, if needed (i.e. runtime
            // deconfiguration, dump/FFDC collection, etc.
            if ( io_sdc.IsUnitCS() && !io_sdc.IsUsingSavedSdc() )
            {
                handleUnitCS( io_sdc, o_dumpTrgt, o_initiateHwudump );
            }

            if ( true == o_initiateHwudump )
            {
                // the dump log will be deleted later in PRDF::main
                // after the hwudump is initiated there.
                o_dumpErrl = iv_errl;
                iv_errl = NULL;
                o_dumpErrlActions = actionFlag;
                PRDF_TRAC( PRDF_FUNC "for target: 0x%08x, i_errl: 0x%08x, "
                           "i_errlActions: 0x%08x", getHuid(o_dumpTrgt),
                           ERRL_GETRC_SAFE(o_dumpErrl), o_dumpErrlActions );
            }
            else
            {
                // Commit the error log.
                commitErrLog( iv_errl, pfaData );
            }
        }
    }

#ifndef __HOSTBOOT_MODULE
    errlHndl_t reg_errl = UtilReg::read ("prdf/RasServices", &sdcSaveFlags, sz_uint8);
    if (reg_errl)
    {
        PRDF_ERR( PRDF_FUNC "Failure in SDC Sync flag Registry read" );
        PRDF_COMMIT_ERRL(reg_errl, ERRL_ACTION_REPORT);
    }
    else
    {
        //Turn off indicator that there is saved Sdc Analysis info
        sdcSaveFlags &= ( ~SDC_ANALYSIS_SAVE_FLAG );
        reg_errl = UtilReg::write ("prdf/RasServices", &sdcSaveFlags, sz_uint8);
        if (reg_errl)
        {
            PRDF_ERR( PRDF_FUNC "Failure in SDC Sync flag Registry write" );
            PRDF_COMMIT_ERRL(reg_errl, ERRL_ACTION_REPORT);
        }
    }
#endif

    PRDF_INF( PRDF_FUNC "PRD called to analyze an error: 0x%08x 0x%08x",
              esig->getChipId(), esig->getSigId() );

    // Reset iv_errl to NULL. This is done to catch logical bug in our code.
    // It enables us to assert in createInitialErrl function if iv_errl is
    // not NULL which should catch any logical bug in initial stages of testing.
    errlHndl_t o_errl = iv_errl;
    iv_errl = NULL;

    return o_errl;

    #undef PRDF_FUNC
}

//------------------------------------------------------------------------------

void ErrDataService::initPfaData( const ServiceDataCollector & i_sdc,
                                  uint32_t i_attnType, bool i_deferDeconfig,
                                  uint32_t i_errlAct, uint32_t i_errlSev,
                                  uint32_t i_prdGardType,
                                  PfaData & o_pfa, TargetHandle_t & o_dumpTrgt )
{
    // Dump info
    o_pfa.msDumpLabel[0] = 0x4D532020; // Start of MS Dump flags
    o_pfa.msDumpLabel[1] = 0x44554D50; // 'MS  DUMP'

    hwTableContent dumpContent;
    i_sdc.GetDumpRequest( dumpContent, o_dumpTrgt );

    // TODO: RTC 116213 proc_mpipl_check_eligibility() is completely broken so
    //       reverting this bug fix to the original 810 and 820 behavior (per
    //       request from Kevin) until proc_mpipl_check_eligibility() is fixed
    //       in FW830.
    //checkMpIplEligibility( i_attnType, o_dumpTrgt, dumpContent );

    o_pfa.msDumpInfo.content = dumpContent;
    o_pfa.msDumpInfo.id      = getHuid(o_dumpTrgt);

    // Error log actions and severity
    o_pfa.errlActions  = i_errlAct;
    o_pfa.errlSeverity = i_errlSev;

    // PRD Service Data Collector Flags (1:true, 0:false)
    o_pfa.DUMP                = i_sdc.IsDump()          ? 1 : 0;
    o_pfa.UERE                = i_sdc.IsUERE()          ? 1 : 0;
    o_pfa.SUE                 = i_sdc.IsSUE()           ? 1 : 0;
    o_pfa.AT_THRESHOLD        = i_sdc.IsAtThreshold()   ? 1 : 0;
    o_pfa.DEGRADED            = i_sdc.IsDegraded()      ? 1 : 0;
    o_pfa.SERVICE_CALL        = i_sdc.IsServiceCall()   ? 1 : 0;
    o_pfa.TRACKIT             = i_sdc.IsMfgTracking()   ? 1 : 0;
    o_pfa.TERMINATE           = i_sdc.Terminate()       ? 1 : 0;
    o_pfa.LOGIT               = i_sdc.IsLogging()       ? 1 : 0;
    o_pfa.UNIT_CHECKSTOP      = i_sdc.IsUnitCS()        ? 1 : 0;
    o_pfa.LAST_CORE_TERMINATE = 0; // Will be set later, if needed.
    o_pfa.USING_SAVED_SDC     = i_sdc.IsUsingSavedSdc() ? 1 : 0;
    o_pfa.DEFER_DECONFIG      = i_deferDeconfig         ? 1 : 0;
    o_pfa.SECONDARY_ERROR     = i_sdc.isSecondaryErrFound() ? 1 : 0;

    // Thresholding
    o_pfa.errorCount = i_sdc.GetHits();
    o_pfa.threshold  = i_sdc.GetThreshold();

    // Misc
    o_pfa.serviceActionCounter = iv_serviceActionCounter;
    o_pfa.prdGardErrType       = i_prdGardType;

    // Attention types
    o_pfa.priAttnType = i_attnType;
    o_pfa.secAttnType = i_sdc.getSecondaryAttnType();

    // Build the MRU list into PFA data.
    const SDC_MRU_LIST & mruList = i_sdc.getMruList();
    uint32_t i; // Iterator used to set limited list count.
    for ( i = 0; i < mruList.size() && i < MruListLIMIT; i++ )
    {
        o_pfa.mruList[i].callout  = mruList[i].callout.flatten();
        o_pfa.mruList[i].type     = mruList[i].callout.getType();
        o_pfa.mruList[i].priority = (uint8_t)mruList[i].priority;

        if( NO_GARD == mruList[i].gardState )
        {
            o_pfa.mruList[i].gardState = GardAction::NoGard;
        }
        else
        {
            o_pfa.mruList[i].gardState = i_prdGardType;
        }
    }
    o_pfa.mruListCount = i;

    // Build the signature list into PFA data
    const PRDF_SIGNATURES & sigList = i_sdc.getSignatureList();
    for ( i = 0; i < sigList.size() && i < SigListLIMIT; i++ )
    {
        o_pfa.sigList[i].chipId    = getHuid(sigList[i].target);
        o_pfa.sigList[i].signature = sigList[i].signature;
    }
    o_pfa.sigListCount = i;

}

//------------------------------------------------------------------------------

void ErrDataService::AddCapData( CaptureData & i_cd, errlHndl_t i_errHdl)
{
    // As CaptureDataClass has large array inside, allocate it on heap
    CaptureDataClass  *l_CapDataBuf = new CaptureDataClass() ;

    for(uint32_t ii = 0; ii < CaptureDataSize; ++ii)
    {
        l_CapDataBuf->CaptureData[ii] = 0xFF;
    }

    uint32_t thisCapDataSize =  i_cd.Copy( l_CapDataBuf->CaptureData,
                                           CaptureDataSize );

    do
    {
        if( 0 == thisCapDataSize )
        {
            // Nothing to add
            break;
        }

        l_CapDataBuf->PfaCaptureDataSize = htonl( thisCapDataSize );

        thisCapDataSize = thisCapDataSize +
                          sizeof(l_CapDataBuf->PfaCaptureDataSize);

        //Compress the Capture data
        size_t l_compressBufSize =
                PrdfCompressBuffer::compressedBufferMax( thisCapDataSize );

        uint8_t * l_compressCapBuf = new uint8_t[l_compressBufSize];

        PrdfCompressBuffer::compressBuffer( ( ( uint8_t * ) l_CapDataBuf ),
                                            (size_t) thisCapDataSize ,
                                            l_compressCapBuf,
                                            l_compressBufSize);

        //Actual size of compressed data is returned in l_compressBufSize
        //Add the Compressed Capture data to Error Log User Data
        PRDF_ADD_FFDC( i_errHdl, (const char*)l_compressCapBuf,
                       (size_t) l_compressBufSize, ErrlVer2, ErrlCapData_1 );

        delete [] l_compressCapBuf;

    }while (0);

    delete l_CapDataBuf;
}

//------------------------------------------------------------------------------

#ifndef __HOSTBOOT_MODULE

bool ErrDataService::SdcSave( sdcSaveFlagsEnum i_saveFlag,
                              ServiceDataCollector & i_saveSdc )
{
    #define PRDF_FUNC "SdcRetrieve() "
    errlHndl_t errorLog = NULL;
    bool rc = false;
    uint8_t sdcSaveFlags = SDC_NO_SAVE_FLAGS;
    size_t  sz_uint8    = sizeof(uint8_t);
    const char * UeSdcKeys[]  = {"fstp/P1_Root","prdf/UeSdcDataPath"};
    const char * SueSdcKeys[] = {"fstp/P1_Root","prdf/SueSdcDataPath"};
    char  * SdcFilename = NULL;
    uint32_t l_size = 0;

    do
    {
        //Need path to the File
        if (i_saveFlag == SDC_SAVE_UE_FLAG)
            errorLog = UtilReg::path(UeSdcKeys,2,NULL,SdcFilename,l_size);
        else if (i_saveFlag == SDC_SAVE_SUE_FLAG)
            errorLog = UtilReg::path(SueSdcKeys,2,NULL,SdcFilename,l_size);
        else
        {
            //Should not get here - code error
            PRDF_ERR( PRDF_FUNC "Failure - incorrect SDC save flag" );
            rc = true;
            break;
        }
        if (errorLog)
        {
            PRDF_ERR( PRDF_FUNC "Failure in getting SDC file path" );
            PRDF_COMMIT_ERRL(errorLog, ERRL_ACTION_REPORT);
            rc = true;
            break;
        }

        rc = SdcWrite(SdcFilename, i_saveSdc);
        if (rc)
        {
            break;
        }

          //Read current SDC state flags from registry
        errorLog = UtilReg::read ("prdf/RasServices", &sdcSaveFlags, sz_uint8);
        if (errorLog)
        {
            PRDF_ERR( PRDF_FUNC "Failure in SDC flag Registry read" );
            PRDF_COMMIT_ERRL(errorLog, ERRL_ACTION_REPORT);
            rc = true;
            break;
        }

        //Update Sdc registry flag
        sdcSaveFlags = (sdcSaveFlags | i_saveFlag);
        errorLog = UtilReg::write ("prdf/RasServices", &sdcSaveFlags, sz_uint8);
        if (errorLog)
        {
            PRDF_ERR( PRDF_FUNC "Failure in SDC flag Registry write" );
            PRDF_COMMIT_ERRL(errorLog, ERRL_ACTION_REPORT);
            rc = true;
            break;
        }


    }
    while(0);

    if (SdcFilename != NULL)
    {  //need to free the pathname
        free(SdcFilename);
        SdcFilename = NULL;
    }

    return rc;

    #undef PRDF_FUNC
}

//------------------------------------------------------------------------------

bool ErrDataService::SdcRetrieve(sdcSaveFlagsEnum i_saveFlag, void * o_buffer)
{
    #define PRDF_FUNC "SdcRetrieve() "
    errlHndl_t errorLog = NULL;
    bool rc = false;
    const char * UeSdcKeys[]  = {"fstp/P1_Root","prdf/UeSdcDataPath"};
    const char * SueSdcKeys[] = {"fstp/P1_Root","prdf/SueSdcDataPath"};
    char  * SdcFilename = NULL;
    uint32_t l_size = 0;

    do
    {
        //Need path to the File
        if (i_saveFlag == SDC_SAVE_UE_FLAG)
            errorLog = UtilReg::path(UeSdcKeys,2,NULL,SdcFilename,l_size);
        else if (i_saveFlag == SDC_SAVE_SUE_FLAG)
            errorLog = UtilReg::path(SueSdcKeys,2,NULL,SdcFilename,l_size);
        else
        {
            //Should not get here - code error
            PRDF_ERR(PRDF_FUNC "Failure - incorrect SDC save flag" );
            rc = true;
            break;
        }
        if (errorLog)
        {
            PRDF_ERR( PRDF_FUNC "Failure in getting SDC file path" );
            PRDF_COMMIT_ERRL(errorLog, ERRL_ACTION_REPORT);
            rc = true;
            break;
        }

        rc = SdcRead (SdcFilename, o_buffer);


    }
    while(0);

    if (SdcFilename != NULL)
    {  //need to free the pathname
        free(SdcFilename);
        SdcFilename = NULL;
    }

    return rc;

    #undef PRDF_FUNC
}

#endif // if not __HOSTBOOT_MODULE

//------------------------------------------------------------------------------

void ErrDataService::deallocateDimms( const SDC_MRU_LIST & i_mruList )
{
    #define PRDF_FUNC "[ErrDataService::deallocateDimms] "

    #if !defined(__HOSTBOOT_MODULE) || defined(__HOSTBOOT_RUNTIME) // RT only

    do
    {
        // First check if Dynamic Memory Deallocation is supported. Then check
        // if it is enabled for predictive callouts.
        if ( !DEALLOC::isEnabled() || !isPredDynDeallocEnabled() ) break;

        TargetHandleList dimmList;
        for ( SDC_MRU_LIST::const_iterator it = i_mruList.begin();
              it != i_mruList.end(); ++it )
        {
            PRDcallout thiscallout = it->callout;
            if ( PRDcalloutData::TYPE_TARGET == thiscallout.getType() )
            {
                TargetHandle_t calloutTgt = thiscallout.getTarget();
                TYPE tgtType = getTargetType( calloutTgt );

                if ( TYPE_L4 == tgtType )
                {
                    calloutTgt = getConnectedParent( calloutTgt, TYPE_MEMBUF );
                    tgtType    = TYPE_MEMBUF;
                }

                if ( (TYPE_MEMBUF == tgtType) ||
                     (TYPE_MBA    == tgtType) ||
                     (TYPE_MCS    == tgtType) )
                {
                    TargetHandleList dimms = getConnected( calloutTgt,
                                                           TYPE_DIMM );
                    dimmList.insert( dimmList.end(), dimms.begin(),
                                     dimms.end() );
                }
                else if ( TYPE_DIMM == tgtType )
                {
                    dimmList.push_back( calloutTgt );
                }
            }
            else if ( PRDcalloutData::TYPE_MEMMRU == thiscallout.getType() )
            {
                MemoryMru memMru (thiscallout.flatten());

                TargetHandleList dimms = memMru.getCalloutList();
                for ( TargetHandleList::iterator dimm = dimms.begin();
                      dimm != dimms.end(); ++dimm )
                {
                    if ( TYPE_DIMM == getTargetType(*dimm) )
                        dimmList.push_back(*dimm);
                }
            }
        }

        if( 0 == dimmList.size() ) break;

        int32_t rc = DEALLOC::dimmListGard( dimmList );
        if ( SUCCESS != rc )
        {
            PRDF_ERR( PRDF_FUNC "dimmListGard failed" );
            break;
        }

    } while(0);

    #endif

    #undef PRDF_FUNC
}

//------------------------------------------------------------------------------
// RasServices class
//------------------------------------------------------------------------------

RasServices::RasServices() :
    iv_eds(NULL)
{
    iv_eds = new ErrDataService();
}

//------------------------------------------------------------------------------

RasServices::~RasServices()
{
    if ( NULL != iv_eds )
    {
        delete iv_eds;
        iv_eds = NULL;
    }
}

//------------------------------------------------------------------------------

void RasServices::Initialize()
{
    iv_eds->Initialize();
}

//------------------------------------------------------------------------------

void RasServices::setErrDataService( ErrDataService & i_eds )
{
    if ( NULL != iv_eds )
    {
        delete iv_eds;
        iv_eds = NULL;
    }

    iv_eds = &i_eds;
}

//------------------------------------------------------------------------------

errlHndl_t RasServices::GenerateSrcPfa( ATTENTION_TYPE i_attnType,
                                        ServiceDataCollector & io_sdc,
                                        bool & o_initiateHwudump,
                                        TargetHandle_t & o_dumpTrgt,
                                        errlHndl_t & o_dumpErrl,
                                        uint32_t & o_dumpErrlActions )

{
    return iv_eds->GenerateSrcPfa( i_attnType, io_sdc, o_initiateHwudump,
                                   o_dumpTrgt, o_dumpErrl, o_dumpErrlActions );
}

//------------------------------------------------------------------------------

void RasServices::createInitialErrl( ATTENTION_TYPE i_attnType )
{
    iv_eds->createInitialErrl( i_attnType );
}

//------------------------------------------------------------------------------

errlHndl_t RasServices::getErrl() const
{
    return iv_eds->getErrl();
}

} // end namespace PRDF

/******************************************************************************/
// Servicability tags for PRDF Ras Services.
// They are located here because their position in the code is not relevant.
/******************************************************************************/

    /*@
      * @errortype
      * @reasoncode PRDF_DETECTED_FAIL_HARDWARE
      * @subsys EPUB_PROCESSOR_SUBSYS
      * @subsys EPUB_PROCESSOR_FRU
      * @subsys EPUB_PROCESSOR_CHIP_CACHE
      * @subsys EPUB_PROCESSOR_UNIT
      * @subsys EPUB_PROCESSOR_BUS_CTL
      * @subsys EPUB_MEMORY_SUBSYS
      * @subsys EPUB_MEMORY_CONTROLLER
      * @subsys EPUB_MEMORY_DIMM
      * @subsys EPUB_MEMORY_FRU
      * @subsys EPUB_EXTERNAL_CACHE
      * @subsys EPUB_CEC_HDW_SUBSYS
      * @subsys EPUB_CEC_HDW_CLK_CTL
      * @subsys EPUB_CEC_HDW_TOD_HDW
      * @subsys EPUB_CEC_HDW_SP_PHYP_INTF
      * @subsys EPUB_MISC_SUBSYS
      * @subsys EPUB_MISC_UNKNOWN
      * @subsys EPUB_MISC_INFORMATIONAL
      * @subsys EPUB_FIRMWARE_SUBSYS
      * @subsys EPUB_FIRMWARE_SP
      * @subsys EPUB_FIRMWARE_PHYP
      * @subsys EPUB_FIRMWARE_HMC
      * @subsys EPUB_EXT_ENVIRO_USER
      * @moduleid PRDF_MAIN
      * @userdata1  PRD Chip Signature
      * @userdata2  PRD Attention Type and Cause Attention Type
      * @userdata3  PRD Signature
      * @devdesc CEC hardware failure detected.
      * @procedure  EPUB_PRC_ALL_PROCS
      * @procedure  EPUB_PRC_REBOOT
      * @procedure  EPUB_PRC_TOD_CLOCK_ERR
      */

    /*@
      * @errortype
      * @reasoncode PRDF_DETECTED_FAIL_HARDWARE_PROBABLE
      * @subsys EPUB_PROCESSOR_SUBSYS
      * @subsys EPUB_PROCESSOR_FRU
      * @subsys EPUB_PROCESSOR_CHIP_CACHE
      * @subsys EPUB_PROCESSOR_UNIT
      * @subsys EPUB_PROCESSOR_BUS_CTL
      * @subsys EPUB_MEMORY_SUBSYS
      * @subsys EPUB_MEMORY_CONTROLLER
      * @subsys EPUB_MEMORY_DIMM
      * @subsys EPUB_MEMORY_FRU
      * @subsys EPUB_EXTERNAL_CACHE
      * @subsys EPUB_CEC_HDW_SUBSYS
      * @subsys EPUB_CEC_HDW_CLK_CTL
      * @subsys EPUB_CEC_HDW_TOD_HDW
      * @subsys EPUB_CEC_HDW_SP_PHYP_INTF
      * @subsys EPUB_MISC_SUBSYS
      * @subsys EPUB_MISC_UNKNOWN
      * @subsys EPUB_MISC_INFORMATIONAL
      * @subsys EPUB_FIRMWARE_SUBSYS
      * @subsys EPUB_FIRMWARE_SP
      * @subsys EPUB_FIRMWARE_PHYP
      * @subsys EPUB_FIRMWARE_HMC
      * @subsys EPUB_EXT_ENVIRO_USER
      * @moduleid PRDF_MAIN
      * @userdata1  PRD Chip Signature
      * @userdata2  PRD Attention Type and Cause Attention Type
      * @userdata3  PRD Signature
      * @devdesc CEC hardware failure detected. Cause is most likely hardware,
      *          but there are other callouts including Software or Next Level
      *          Support.
      * @procedure EPUB_PRC_SP_CODE
      * @procedure  EPUB_PRC_PHYP_CODE
      * @procedure  EPUB_PRC_LVL_SUPP
      * @procedure  EPUB_PRC_ALL_PROCS
      * @procedure  EPUB_PRC_REBOOT
      * @procedure  EPUB_PRC_TOD_CLOCK_ERR
      */

    /*@
      * @errortype
      * @reasoncode PRDF_DETECTED_FAIL_SOFTWARE_PROBABLE
      * @subsys EPUB_PROCESSOR_SUBSYS
      * @subsys EPUB_PROCESSOR_FRU
      * @subsys EPUB_PROCESSOR_CHIP_CACHE
      * @subsys EPUB_PROCESSOR_UNIT
      * @subsys EPUB_PROCESSOR_BUS_CTL
      * @subsys EPUB_MEMORY_SUBSYS
      * @subsys EPUB_MEMORY_CONTROLLER
      * @subsys EPUB_MEMORY_DIMM
      * @subsys EPUB_MEMORY_FRU
      * @subsys EPUB_EXTERNAL_CACHE
      * @subsys EPUB_CEC_HDW_SUBSYS
      * @subsys EPUB_CEC_HDW_CLK_CTL
      * @subsys EPUB_CEC_HDW_TOD_HDW
      * @subsys EPUB_CEC_HDW_SP_PHYP_INTF
      * @subsys EPUB_MISC_SUBSYS
      * @subsys EPUB_MISC_UNKNOWN
      * @subsys EPUB_MISC_INFORMATIONAL
      * @subsys EPUB_FIRMWARE_SUBSYS
      * @subsys EPUB_FIRMWARE_SP
      * @subsys EPUB_FIRMWARE_PHYP
      * @subsys EPUB_FIRMWARE_HMC
      * @subsys EPUB_EXT_ENVIRO_USER
      * @moduleid PRDF_MAIN
      * @userdata1  PRD Chip Signature
      * @userdata2  PRD Attention Type and Cause Attention Type
      * @userdata3  PRD Signature
      * @devdesc CEC hardware failure detected. Cause is most likley Software,
      *          but there are also Hardware callouts.
      * @procedure EPUB_PRC_SP_CODE
      * @procedure  EPUB_PRC_PHYP_CODE
      * @procedure  EPUB_PRC_LVL_SUPP
      * @procedure  EPUB_PRC_ALL_PROCS
      * @procedure  EPUB_PRC_REBOOT
      * @procedure  EPUB_PRC_TOD_CLOCK_ERR
      */

    /*@
      * @errortype
      * @reasoncode PRDF_DETECTED_FAIL_SOFTWARE
      * @subsys EPUB_MISC_SUBSYS
      * @subsys EPUB_MISC_UNKNOWN
      * @subsys EPUB_FIRMWARE_SUBSYS
      * @subsys EPUB_FIRMWARE_SP
      * @subsys EPUB_FIRMWARE_PHYP
      * @subsys EPUB_FIRMWARE_HMC
      * @moduleid PRDF_MAIN
      * @userdata1  PRD Chip Signature
      * @userdata2  PRD Attention Type and Cause Attention Type
      * @userdata3  PRD Signature
      * @devdesc CEC hardware failure detected. Cause is most likely Software.
      * @procedure EPUB_PRC_SP_CODE
      * @procedure  EPUB_PRC_PHYP_CODE
      * @procedure  EPUB_PRC_LVL_SUPP
      * @procedure  EPUB_PRC_ALL_PROCS
      * @procedure  EPUB_PRC_REBOOT
      * @procedure  EPUB_PRC_TOD_CLOCK_ERR
      */

